#  -*- coding: utf-8 -*-
# (c) 2018 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import logging
from datetime import datetime

from grafanimate.animations import SequentialAnimation

logger = logging.getLogger(__name__)


class AnimationScenario:
    """
    Run different anmiation scenarios/sequences.

    As the ad-hoc interface is not finished yet,
    this is all we got. Enjoy!
    """

    def __init__(self, grafana, storage, dashboard_uid=None, target=None, options=None):

        self.grafana = grafana
        self.storage = storage
        self.dashboard_uid = dashboard_uid

        # Dispatch output to different target formats.
        # FIXME: Not implemented yet.
        self.target = target

        # Start the engines.
        self.engine = SequentialAnimation(grafana=grafana, dashboard_uid=dashboard_uid, options=options)
        self.engine.start()

    def ldi_all(self):
        logger.info('Running scenario ldi_all')

        # LDI, ramp-up
        results = self.engine.run(dtstart=datetime(2015, 10, 1), dtuntil=datetime(2017, 1, 1), interval='monthly')
        self.storage.save_items(results)

        # LDI, growth
        results = self.engine.run(dtstart=datetime(2017, 1, 1), dtuntil=datetime.now(), interval='weekly')
        self.storage.save_items(results)

        self.storage.save_items(results)

    def ldi_with_gaps(self):
        logger.info('Running scenario ldi_with_gaps')

        # LDI, ramp-up
        results = self.engine.run(dtstart=datetime(2015, 10, 1), dtuntil=datetime(2017, 1, 1), interval='monthly')
        self.storage.save_items(results)

        # LDI, growth, with gap at 2018-04-29 - 2018-12-20
        # TODO: Detect empty data from datasource through Grafana Sidecar and skip respective images.
        results = self.engine.run(dtstart=datetime(2017, 1, 1), dtuntil=datetime(2018, 6, 5), interval='weekly')
        self.storage.save_items(results)

        results = self.engine.run(dtstart=datetime(2018, 12, 20), dtuntil=datetime.now(), interval='weekly')
        self.storage.save_items(results)

    def cdc_maps(self):
        logger.info('Running scenario cdc_maps')

        # CDC, temperatur-sonne-and-niederschlag-karten
        results = self.engine.run(dtstart=datetime(2018, 3, 6, 5, 0, 0), dtuntil=datetime(2018, 3, 10, 23, 59, 59), interval='hourly')

        # Short sequence, for debugging processes.
        #results = self.engine.run(dtstart=datetime(2018, 3, 6, 5, 0, 0), dtuntil=datetime(2018, 3, 6, 6, 59, 59), interval='hourly')

        self.storage.save_items(results)

    def ir_sensor_svg_pixmap(self):
        """
        dtstart: 2018-08-14 03:16:00
        dtuntil: 2018-08-14 03:16:36
        """
        logger.info('Running scenario ir_sensor_svg_pixmap')

        results = self.engine.run(dtstart=datetime(2018, 8, 14, 3, 16, 0), dtuntil=datetime(2018, 8, 14, 3, 16, 36), interval='secondly')
        self.storage.save_items(results)



# -*- coding: utf-8 -*-
# (c) 2018 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
from grafanimate import postprocessing
from grafanimate.scenarios import logger
from grafanimate.util import format_date_human, slug, ensure_directory


class MediaStorage:

    def __init__(self, imagefile, outputfile):
        self.imagefile_template = imagefile
        self.outputfile_template = outputfile

    def save_items(self, results):

        for item in results:
            #logger.info('item: %s', item)
            printable = item.copy()
            printable.data.image = printable.data.image[:23] + '...'
            logger.debug('Item: %s', printable)
            self.save_item(item)

    def save_item(self, item):

        # Compute image sequence file name.
        imagefile = self.imagefile_template.format(
            interval=item.meta.interval,
            uid=item.meta.dashboard,
            date=format_date_human(item.data.dtstart))

        # Ensure directory exists.
        ensure_directory(imagefile)

        # Store image.
        with open(imagefile, 'w') as f:
            f.write(item.data.image)

        logger.info('Saved frame to {}. Size: {}'.format(imagefile, len(item.data.image)))

    def produce_artifacts(self, uid, name=None):
        # TODO: Can use dashboard title as output filename here?
        # TODO: Can put dtstart into filename?

        #uid = self.dashboard_uid
        name = name or uid

        # Compute path to sequence images.
        imagefile_pattern = self.imagefile_template.format(uid=uid, date='*')

        # Compute output file name.
        if name:
            name = slug(name)
        outputfile = self.outputfile_template.format(name=name)

        # Produce output artifacts.
        ensure_directory(outputfile)
        return postprocessing.render(imagefile_pattern, outputfile)



#  -*- coding: utf-8 -*-
# (c) 2018 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import logging
from munch import munchify
from datetime import timedelta
from dateutil.relativedelta import relativedelta
from dateutil.rrule import rrule, SECONDLY, MINUTELY, HOURLY, DAILY, WEEKLY, MONTHLY, YEARLY

logger = logging.getLogger(__name__)


class SequentialAnimation:

    def __init__(self, grafana, dashboard_uid=None, options=None):

        self.grafana = grafana
        self.dashboard_uid = dashboard_uid
        self.options = options or None

    def start(self):
        logger.info('Opening dashboard')
        self.grafana.open_dashboard(self.dashboard_uid, options=self.options)
        logger.info('Dashboard ready')

    def run(self, dtstart=None, dtuntil=None, interval=None):

        if dtstart > dtuntil:
            message = 'Timestamp dtstart={} is after dtuntil={}'.format(dtstart, dtuntil)
            raise ValueError(message)

        #flavor = 'expand'
        flavor = 'window'

        rr_freq, rr_interval, dtdelta = self.get_freq_delta(interval)

        #until = datetime.now()
        if flavor == 'expand':
            dtuntil += dtdelta

        # Compute complete date range.
        logger.info('Creating rrule: dtstart=%s, until=%s, freq=%s, interval=%s', dtstart, dtuntil, rr_freq, rr_interval)
        daterange = list(rrule(dtstart=dtstart, until=dtuntil, freq=rr_freq, interval=rr_interval))
        #logger.info('Date range is: %s', daterange)

        # Iterate date range.
        for date in daterange:

            logger.info('Datetime step: %s', date)

            # Compute start and end dates based on flavor.

            if flavor == 'window':
                dtstart = date
                dtuntil = date + dtdelta

            elif flavor == 'expand':
                dtuntil = date

            # Render image.
            image = self.render(dtstart, dtuntil, interval)

            # Build item model.
            item = munchify({
                'meta': {
                    'grafana': self.grafana,
                    'dashboard': self.dashboard_uid,
                    'interval': interval,
                },
                'data': {
                    'dtstart': dtstart,
                    'dtuntil': dtuntil,
                    'image': image,
                },
            })

            yield item

    def get_freq_delta(self, interval):

        rr_interval = 1

        # Secondly
        if interval == 'secondly':
            rr_freq = SECONDLY
            delta = timedelta(seconds=1)

        # Minutely
        elif interval == 'minutely':
            rr_freq = MINUTELY
            delta = timedelta(minutes=1) - timedelta(seconds=1)

        # Each 10 minutes
        elif interval == '10min':
            rr_freq = MINUTELY
            rr_interval = 10
            delta = timedelta(minutes=10) - timedelta(seconds=1)

        # Hourly
        elif interval == 'hourly':
            rr_freq = HOURLY
            delta = timedelta(hours=1) - timedelta(seconds=1)

        # Daily
        elif interval == 'daily':
            rr_freq = DAILY
            delta = timedelta(days=1) - timedelta(seconds=1)

        # Weekly
        elif interval == 'weekly':
            rr_freq = WEEKLY
            delta = timedelta(weeks=1) - timedelta(seconds=1)

        # Monthly
        elif interval == 'monthly':
            rr_freq = MONTHLY
            delta = relativedelta(months=+1) - relativedelta(seconds=1)

        # Yearly
        elif interval == 'yearly':
            rr_freq = YEARLY
            delta = relativedelta(years=+1) - relativedelta(seconds=1)

        else:
            raise ValueError('Unknown interval "{}"'.format(interval))

        return rr_freq, rr_interval, delta

    def render(self, dtstart, dtuntil, interval):

        logger.debug('Adjusting time range control')
        self.grafana.timewarp(dtstart, dtuntil, interval)

        logger.debug('Rendering image')
        return self.make_image()

    def make_image(self):
        image = self.grafana.render_image()
        #logger.info('Image size: %s', len(image))
        return image



#  -*- coding: utf-8 -*-
# (c) 2018 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import json
import logging
from pkg_resources import resource_stream
from marionette_driver import Wait
from marionette_driver.errors import TimeoutException
from grafanimate.marionette import FirefoxMarionetteBase
from grafanimate.util import format_date_grafana

log = logging.getLogger(__name__)


class GrafanaWrapper(FirefoxMarionetteBase):
    """
    https://marionette-client.readthedocs.io/en/master/interactive.html
    """

    def __init__(self, baseurl=None):
        self.baseurl = baseurl
        log.info('Starting GrafanaWrapper on %s', baseurl)
        FirefoxMarionetteBase.__init__(self)

    def boot_grafana(self):
        """
        Navigate to Grafana application and inject Grafana Sidecar service.
        """
        log.info('Starting Grafana at {}'.format(self.baseurl))

        self.fix_window_size()

        # Navigate to resource URL.
        self.marionette.navigate(self.baseurl)

        # Wait for Grafana application to load.
        self.wait_for_grafana()

        # Load Javascript Grafana Sidecar service.
        with resource_stream('grafanimate', 'grafana-studio.js') as f:
            javascript = f.read()
            self.run_javascript(javascript)

    def fix_window_size(self):
        """
        Work around ffmpeg errors like::

          Input #0, image2, from './var/spool/DLOlE_Rmz/DLOlE_Rmz_*.png':
            Duration: 00:00:28.75, start: 0.000000, bitrate: N/A
              Stream #0:0: Video: png, rgba(pc), 1497x483, 4 fps, 4 tbr, 4 tbn, 4 tbc

        [libx264 @ 0x7fcf0c001200] width not divisible by 2 (1497x483)

        [libx264 @ 0x7fa917001200] height not divisible by 2 (1348x823)
        """
        window_size = self.get_window_rect()
        if window_size['width'] % 2:
            window_size['width'] -= 1
        if window_size['height'] % 2:
            window_size['height'] -= 1
        self.set_window_size(window_size['width'], window_size['height'])

    def wait_for_grafana(self):
        """
        Wait for element <grafana-app> to appear.
        """
        log.info('Waiting for Grafana to load')
        element = self.wait_for_element_tag("grafana-app")
        log.info('Grafana loaded')
        return element

    def open_dashboard(self, uid, options=None):
        """
        Open a Grafana Dashboard with the designated uid
        and wait for all data to load into all panels.
        """
        options = options or {}
        javascript = mkjscall("grafanaStudio.openDashboard", uid, options)
        log.info('Running Javascript: %s', javascript)
        self.run_javascript(javascript)
        self.wait_all_data_received()

    def get_dashboard_title(self):
        return self.calljs("grafanaStudio.getDashboardTitle")

    def wait_all_data_received(self):
        """
        Wait for all data to arrive in the dashboard.
        """

        log.info('Waiting for "all-data-received" event')
        waiter = Wait(self.marionette, timeout=20.0, interval=1.0)

        def condition(marionette):
            return self.calljs("grafanaStudio.hasAllData")

        try:
            waiter.until(condition)
        except TimeoutException as ex:
            log.warning('Timed out waiting for data: %s. Continuing anyway.', ex)

    def clear_all_data_received(self):
        return self.calljs("grafanaStudio.hasAllData", False)

    def timewarp(self, dtstart, dtuntil, interval):
        """
        Navigate the Dashboard to the designated point in time
        and wait for refreshing all child components including data.
        """

        # Notify user.
        message = 'Timewarp to {} -> {}'.format(dtstart, dtuntil)
        log.info(message)
        self.console_log(message)

        # Perform timewarp.
        self.clear_all_data_received()
        self.timerange_set(format_date_grafana(dtstart, interval), format_date_grafana(dtuntil, interval))
        self.wait_all_data_received()

    def timerange_set(self, starttime, endtime):
        """
        Adjust Grafana time control. This is not synchronous.
        """
        return self.calljs("grafanaStudio.setTime", starttime, endtime)

    def timerange_get(self):
        """
        For fetching the current timeRange values, use::

            var timeRange = angular.element('grafana-app').injector().get('timeSrv').timeRange();
            var temp_date_from = new Date(timeRange.from);
            var temp_date_to = new Date(timeRange.to);

        -- https://community.grafana.com/t/how-to-access-time-picker-from-to-within-a-text-panel-and-jquery/6071/3
        """
        return self.calljs("grafanaStudio.getTime")
        raise NotImplemented("timerange_get not implemented yet")

    def run_javascript(self, sourcecode):
        """
        Run the designated Javascript code directly in the scope
        of the application under test. Don't use any sandboxing.

        :code: Plain Javascript source code.
        """
        # https://github.com/mozilla/geckodriver/issues/1067#issuecomment-347180427
        # https://github.com/devtools-html/har-export-trigger/issues/27#issuecomment-424777524
        log.debug('Running Javascript: %s', sourcecode)
        return self.marionette.execute_script(sourcecode, sandbox=None, new_sandbox=False)

    def calljs(self, name, *args):
        return self.run_javascript(mkjscall(name, *args))

    def console_log(self, message):
        """
        Write a message to the Browser console.
        """
        return self.calljs("console.log", message)

    def wait_for_map(self):
        """
        Wait for map element <panel-plugin-grafana-worldmap-panel ...> to appear.
        """
        log.debug('Waiting for map element to appear in DOM')
        element = self.wait_for_element_tag("panel-plugin-grafana-worldmap-panel")
        log.info('Finished waiting for map element')
        log.info(element)
        return element


def mkjscall(name, *arguments, **flags):
    flags.setdefault('add_return', True)
    tpl = '{return}{name}({arguments});'
    tplvars = {
        'return': flags['add_return'] and 'return ' or '',
        'name': name,
        'arguments': mkjsargs(*arguments),
    }
    return tpl.format(**tplvars)


def mkjsargs(*arguments):
    return json.dumps(arguments).strip('[]')




# -*- coding: utf-8 -*-
# (c) 2018 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import json
import logging
from docopt import docopt, DocoptExit
from grafanimate import __appname__, __version__
from grafanimate.core import make_grafana, make_animation, make_storage
from grafanimate.util import normalize_options, setup_logging

log = logging.getLogger(__name__)


def run():
    """
    Usage:
      grafanimate [options] [--target=<target>]...
      grafanimate --version
      grafanimate (-h | --help)

    Options:
      --grafana-url=<url>           Base URL to Grafana, [default: http://localhost:3000].
      --scenario=<scenario>         Which scenario to run. Scenarios are defined as methods.
      --dashboard-uid=<uid>         Grafana dashboard uid.

    Optional:
      --panel-id=<id>               Render single panel only by navigating to "panelId=<id>&fullscreen".
      --dashboard-view=<mode>       Use Grafana's "d-solo" view for rendering single panels without header.

      --header-layout=<layout>      The header rendering subsystem offers different modes
                                    for amending the vanilla Grafana user interface.
                                    Multiple modes can be combined.
                                    [default: large-font]

                                    - no-chrome:            Set kiosk mode, remove sidemenu and more chrome
                                    - large-font:           Use larger font sizes for title and datetime
                                    - collapse-datetime:    Collapse datetime into title
                                    - studio:               Apply studio modifications. This options aggregates
                                                            "no-chrome", "large-font" and "collapse-datetime".
                                    - no-folder:            Don't include foldername in title

                                    - no-title:             Turn off title widget
                                    - no-datetime:          Turn off datetime widget

      --datetime-format=<format>    Datetime format to use with header layouts like "studio".
                                    Examples: YYYY-MM-DD HH:mm:ss, YYYY, HH:mm.

                                    There are also some format presets available here:
                                    - human-date:           on 2018-08-14
                                    - human-time:           at 03:16:05
                                    - human-datetime:       on 2018-08-14 at 03:16:05

                                    When left empty, the default is determined by the configured interval.

      --debug                       Enable debug logging
      -h --help                     Show this screen


    Examples for scenario mode. Script your animation in file "scenarios.py".

      # Generate sequence of .png files in ./var/spool/ldi_all/1aOmc1sik
      grafanimate --grafana-url=http://localhost:3000/ --scenario=ldi_all --dashboard-uid=1aOmc1sik

      # Use more parameters to control rendering process.
      # grafanimate \\
      #   --grafana-url=http://localhost:3000/ --scenario=ir_sensor_svg_pixmap --dashboard-uid=_TbvFUyik \\
      #   --header-layout=studio --datetime-format=human-time --panel-id=6


    NOT IMPLEMENTED YET

      --target=<target>             Data output target

    Examples: Ad hoc mode.

    Until implemented, please use scenario mode.
    Don't be afraid, it's just some copy/pasting in the "scenarios.py" file, go ahead.

      --start=<start>               Start time
      --end=<end>                   End time
      --interval=<end>              Interval time

    """

    # Parse command line arguments.
    options = docopt(run.__doc__, version=__appname__ + ' ' + __version__)
    options = normalize_options(options, lists=['header-layout'])

    # Setup logging.
    debug = options.get('debug')
    log_level = logging.INFO
    if debug:
        log_level = logging.DEBUG
    setup_logging(log_level)

    # Debug command line options.
    #import json; log.info('Options: {}'.format(json.dumps(options, indent=4)))

    # Sanity checks.
    if not options['scenario']:
        raise DocoptExit('Error: Parameter --scenario is mandatory')

    if not options['dashboard-uid']:
        raise DocoptExit('Error: Parameter --dashboard-uid is mandatory')

    if options['dashboard-view'] == 'd-solo' and not options['panel-id']:
        raise DocoptExit('Error: Parameter --panel-id is mandatory for --dashboard-view=d-solo')

    # Define and run Pipeline.

    # Define pipeline elements.
    grafana = make_grafana(options['grafana-url'])
    storage = make_storage(
        imagefile='./var/spool/{uid}/{uid}_{date}.png',
        outputfile='./var/results/{name}.mp4')

    # Assemble pipeline.
    animation = make_animation(grafana, storage, options)

    # Run stop motion animation to produce single artifacts.
    animation.run()

    # Run rendering steps, produce composite artifacts.
    title = grafana.get_dashboard_title()
    results = storage.produce_artifacts(options['dashboard-uid'], title)

    log.info('Produced %s results\n%s', len(results), json.dumps(results, indent=2))



#  -*- coding: utf-8 -*-
# (c) 2018 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import json
import where
import atexit
import logging
from collections import OrderedDict

from marionette_driver import Wait
from marionette_driver.marionette import Marionette
from marionette_driver.errors import NoSuchElementException

from grafanimate.util import check_socket, find_program_candidate

logger = logging.getLogger(__name__)


class FirefoxMarionetteBase(object):
    """
    Wrap Marionette/Firefox into convenient interface.

    - https://marionette-client.readthedocs.io/
    - https://marionette-client.readthedocs.io/en/master/reference.html
    - https://marionette-client.readthedocs.io/en/master/interactive.html
    """
    def __init__(self):
        logger.info('Starting Marionette Gecko wrapper')

        # Configuration
        self.firefox_bin = self.find_firefox()
        self.firefox_host = 'localhost'
        self.firefox_port = 2828
        # TODO: Make configurable
        self.firefox_verbosity = 1
        #self.firefox_verbosity = 2

        # Timeout configuration
        self.startup_timeout = 20.0
        self.socket_timeout = 32.0
        self.page_timeout = 30.0
        self.script_timeout = 20.0
        self.shutdown_timeout = 10.0

        # Instance state defaults
        self.marionette = None
        self.firefox_run_headless = True
        self.firefox_do_shutdown = False
        self.firefox_already_started = False

    def enable_headless(self, run_headless=True):
        self.firefox_run_headless = run_headless

    def enable_shutdown(self, do_shutdown=True):
        self.firefox_do_shutdown = do_shutdown

    def boot_firefox(self, headless=True):

        # Indicate whether to run in headless mode
        self.enable_headless(headless)

        # Optionally shut down Marionette/Firefox after performing work
        # This will just be called if Python exits normally
        atexit.register(self.shutdown)

        # Check whether Firefox is already running
        logger.info('Check for running instance of Marionette/Firefox at {}:{}'.format(
            self.firefox_host, self.firefox_port))

        if check_socket(self.firefox_host, self.firefox_port):
            logger.info('Will reuse running Marionette/Firefox')
            self.firefox_bin = None
            self.firefox_already_started = True
        else:
            logger.info('Will launch new Marionette/Firefox instance')

        # Connect to / start Marionette Gecko engine
        self.marionette = Marionette(
            host=self.firefox_host, port=self.firefox_port, bin=self.firefox_bin,
            socket_timeout=self.socket_timeout, startup_timeout=self.startup_timeout,
            headless=self.firefox_run_headless, verbose=self.firefox_verbosity)

        self.marionette.DEFAULT_SHUTDOWN_TIMEOUT = self.shutdown_timeout

        # Start a session with Marionette Gecko engine
        self.marionette.start_session()

        # Configure Marionette
        self.configure_marionette()

    def configure_marionette(self):

        # This specifies the time to wait for the page loading to complete.
        self.marionette.timeout.page_load = self.page_timeout

        # This specifies the time to wait for injected scripts to finish
        # before interrupting them.
        self.marionette.timeout.script = self.script_timeout

        # Configure a HTTP proxy server
        self.marionette.set_pref('network.proxy.type', 0, default_branch=True)

    @classmethod
    def find_firefox(cls):
        candidates = where.where('firefox')
        candidates += [
            '/Applications/Firefox.app/Contents/MacOS/firefox-bin',
        ]
        firefox = find_program_candidate(candidates)
        logger.info('Found "firefox" program at {}'.format(firefox))
        return firefox

    def get_status(self):
        attributes = ['session', 'session_id']
        data = OrderedDict()
        for attribute in attributes:
            data[attribute] = getattr(self.marionette, attribute)
        return data

    def log_status(self):
        logger.info('Marionette report: {}'.format(json.dumps(self.get_status(), indent=4)))

    def has_active_session(self):
        is_initialized = self.marionette is not None and self.marionette.session_id is not None
        return is_initialized

    def ensure_session(self):
        #self.log_status()
        if not self.has_active_session():
            self.boot_firefox()
            logger.info('No session with Marionette, started new session {}'.format(self.marionette.session_id))

    def shutdown(self):
        if self.firefox_do_shutdown:

            logger.info('Aiming at shutdown')

            if self.firefox_already_started:
                logger.warning('Can not shutdown Firefox as it was already running before starting this program')
                return False

            logger.info('Shutting down Marionette/Firefox')
            if self.marionette is not None:
                self.marionette.quit()
                return True

    def find_tag(self, tagname):
        try:
            element = self.marionette.find_element("tag name", tagname)
            return element
        except NoSuchElementException:
            pass

    def wait_for_element_tag(self, tagname):
        """
        Wait for element to appear.
        """
        waiter = Wait(self.marionette, timeout=20.0, interval=0.1)
        element = waiter.until(lambda m: self.find_tag(tagname))
        return element

    def render_image(self, element=None):
        """
        Return screenshot from element.
        """
        image = self.marionette.screenshot(element=element, format='binary')
        return image

    def set_window_size(self, width, height):
        self.marionette.set_window_rect(width=width, height=height)

    def get_window_rect(self):
        return self.marionette.window_rect



# -*- coding: utf-8 -*-
# (c) 2018 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import logging

from grafanimate.grafana import GrafanaWrapper
from grafanimate.scenarios import AnimationScenario
from grafanimate.mediastorage import MediaStorage
from grafanimate.util import filter_dict

log = logging.getLogger(__name__)


def make_grafana(url):
    grafana = GrafanaWrapper(baseurl=url)
    grafana.boot_firefox(headless=False)
    grafana.boot_grafana()
    return grafana


def make_storage(imagefile=None, outputfile=None):
    return MediaStorage(imagefile=imagefile, outputfile=outputfile)


def make_animation(grafana, storage, options):

    # Prepare animation.
    scenario = AnimationScenario(
        grafana=grafana,
        storage=storage,
        dashboard_uid=options['dashboard-uid'],
        options=filter_dict(options, ['panel-id', 'dashboard-view', 'header-layout', 'datetime-format'])
    )
    if not hasattr(scenario, options.scenario):
        raise NotImplementedError('Animation scenario "{}" not implemented'.format(options.scenario))

    # Run animation scenario.
    scenario.run = getattr(scenario, options.scenario)
    return scenario


    # TODO: Introduce ad-hoc mode. In the meanwhile, please use scenario mode.
    """
    animator = SequentialAnimation(
        options['url'],
        time_start=options.get('start'),
        time_end=options.get('end', 'now'),
        time_step=options.get('interval', '1h')
    )
    animator.run()
    """



#  -*- coding: utf-8 -*-
# (c) 2018 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import os
import logging

logger = logging.getLogger(__name__)


def to_video(source, target):
    """
    http://hamelot.io/visualization/using-ffmpeg-to-convert-a-set-of-images-into-a-video/
    https://stackoverflow.com/questions/24961127/how-to-create-a-video-from-images-with-ffmpeg

    https://stackoverflow.com/questions/38726370/ffmpeg-text-watermark-bottom-left/38728172#38728172
    https://superuser.com/questions/939357/ffmpeg-watermark-on-bottom-right-corner/939386#939386
    """
    #command = "ffmpeg -framerate 4 -pattern_type glob -i '{}' -c:v libx264 -r 30 -pix_fmt yuv420p '{}' -y".format(source, target)
    # -vf "fps=25,format=yuv420p,drawtext=fontfile=OpenSans-Regular.ttf:text='Title of this Video':fontcolor=white:fontsize=24:x=(w-tw)/2:y=(h/PHI)+th
    #command = "ffmpeg -framerate 4 -pattern_type glob -i '{}' -c:v libx264 -vf 'fps=25,format=yuv420p,drawtext=text=Produced with Grafana and grafanimate:fontsize=11:x=w-tw-30:y=h-th-10:fontcolor=lightgrey:fontfile=/Library/Fonts/Arial.ttf' '{}' -y".format(source, target)

    command = "ffmpeg -framerate 4 -pattern_type glob -i '{}' -c:v libx264 -vf 'fps=25,format=yuv420p' '{}' -y".format(source, target)
    logger.info('Rendering video: {}'.format(target))
    logger.debug(command)
    os.system(command)


def to_gif(source, target):
    """
    # High Quality Gifs with FFMPEG
    # https://medium.com/@colten_jackson/doing-the-gif-thing-on-debian-82b9760a8483
    ::

        ffmpeg -ss 2.6 -t 1.3 -i MVI_7035.MOV \
            -vf fps=15,scale=320:-1:flags=lanczos,palettegen palette.png

        ffmpeg -ss 2.6 -t 1.3 -i MVI_7035.MOV -i palette.png \
            -filter_complex “fps=15,scale=400:-1:flags=lanczos[x];[x][1:v]paletteuse” sixthtry.gif

    > I think the -1 in the video filters refers to the height and basically means ‘preserve aspect ratio’.
    > So, in this case, 320:-1 means scale to: w=320, h=320/CurrentWidth*CurrentHeight

    # How to make GIFs with FFMPEG
    https://engineering.giphy.com/how-to-make-gifs-with-ffmpeg/

    # How do I convert a video to GIF using ffmpeg, with reasonable quality?
    https://superuser.com/questions/556029/how-do-i-convert-a-video-to-gif-using-ffmpeg-with-reasonable-quality

    # High quality GIF with FFmpeg
    http://blog.pkh.me/p/21-high-quality-gif-with-ffmpeg.html

    # FFMPEG gif script for bash
    https://github.com/thevangelist/FFMPEG-gif-script-for-bash

    > Here is a better version with fifo filter to avoid Buffer queue overflow when using paletteuse filter.
    > By using split filter to avoid the creation of intermediate palette PNG file.
    > -- https://superuser.com/questions/556029/how-do-i-convert-a-video-to-gif-using-ffmpeg-with-reasonable-quality/1256459#1256459

    ffmpeg -i input.mp4 -filter_complex 'fps=10,scale=320:-1:flags=lanczos,split [o1] [o2];[o1] palettegen [p]; [o2] fifo [o3];[o3] [p] paletteuse' out.gif
    """

    """
    # PoC
    ffmpeg -i dwd-cdc-2018-08.mov -vf fps=15,scale=320:-1:flags=lanczos,palettegen palette.png
    ffmpeg -i dwd-cdc-2018-08.mov -i palette.png -filter_complex "fps=15,scale=480:-1:flags=lanczos[x];[x][1:v]paletteuse" dwd-cdc-2018-08.gif -y
    """

    """
    # One-step conversion.
    ffmpeg -i dwd-cdc-2018-08.mov -filter_complex 'fps=10,scale=480:-1:flags=lanczos,split [o1] [o2];[o1] palettegen [p]; [o2] fifo [o3];[o3] [p] paletteuse' dwd-cdc-2018-08-v2.gif -y
    """

    command = "ffmpeg -i '{}' -filter_complex 'fps=10,scale=480:-1:flags=lanczos,split [o1] [o2];[o1] palettegen [p]; [o2] fifo [o3];[o3] [p] paletteuse' '{}' -y".format(source, target)
    logger.info('Rendering GIF: {}'.format(target))
    logger.debug(command)
    os.system(command)


def upload_server(source):
    command = "make --makefile=/Users/amo/dev/hiveeyes/sources/documentation/Makefile ptrace source={}".format(source)
    os.system(command)


def render(source, target):
    mp4 = target
    suffix = '.' + target.split('.')[-1]
    gif = mp4.replace(suffix, '.gif')
    to_video(source, mp4)
    to_gif(mp4, gif)
    results = [mp4, gif]
    return results


def run(source, target):
    render(source, target)
    #upload_server(gif)
    #upload_server(mp4)


if __name__ == '__main__':
    run('./var/spool/*_1aOmc1sik_*.png', './var/results/ldi-coverage.mp4')

    #to_video('./var/spool/*_1aOmc1sik_*.png', 'ldi-coverage.mp4')
    #to_gif('ldi-coverage.mp4', 'ldi-coverage.gif')
    #upload_server('ldi-coverage.gif')
    #upload_server('ldi-coverage.mp4')



#  -*- coding: utf-8 -*-
# (c) 2018 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import re
import os
import sys
import socket
import logging
from munch import munchify
from contextlib import closing
from unidecode import unidecode


def setup_logging(level=logging.INFO):
    log_format = '%(asctime)-10s [%(name)-30s] %(levelname)-7s: %(message)s'
    logging.basicConfig(
        format=log_format,
        stream=sys.stderr,
        level=level)

    # TODO: Control debug logging of HTTP requests through yet another commandline option "--debug-http" or "--debug-requests"
    requests_log = logging.getLogger('requests')
    requests_log.setLevel(logging.WARN)


def normalize_options(options, lists=None):
    lists = lists or []
    normalized = {}
    for key, value in options.items():
        key = key.strip('--<>')
        normalized[key] = value
    for key in lists:
        normalized[key] = read_list(normalized[key])
    return munchify(normalized)


def read_list(data, separator=u','):
    if data is None:
        return []
    result = list(map(lambda x: x.strip(), data.split(separator)))
    if len(result) == 1 and not result[0]:
        result = []
    return result


def find_program_candidate(candidates):
    for candidate in candidates:
        if os.path.isfile(candidate):
            return candidate


def check_socket(host, port):
    # https://stackoverflow.com/questions/19196105/python-how-to-check-if-a-network-port-is-open-on-linux
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        sock.settimeout(1)
        if sock.connect_ex((host, port)) == 0:
            return True
        else:
            return False


def format_date_human(date, interval=None):
    #pattern = '%Y-%m-%d'
    pattern = '%Y-%m-%dT%H-%M-%S'
    #if interval in ['secondly', 'minutely', 'hourly']:
    #    pattern = '%Y-%m-%dT%H-%M-%S'
    date_formatted = date.strftime(pattern)
    return date_formatted


def format_date_grafana(date, interval=None):
    pattern = '%Y-%m-%d'
    if interval in ['secondly', 'minutely', 'hourly'] or interval.endswith('min'):
        pattern = '%Y-%m-%dT%H:%M:%S'
    date_formatted = date.strftime(pattern)
    return date_formatted


def filter_dict(data, keys):
    return {k: v for k, v in data.items() if k in keys}


def slug(text):
    """Make a URL-safe, human-readable version of the given text

    This will do the following:

    1. decode unicode characters into ASCII
    2. shift everything to lowercase
    3. strip whitespace
    4. replace other non-word characters with dashes
    5. strip extra dashes

    This somewhat duplicates the :func:`Google.slugify` function but
    slugify is not as generic as this one, which can be reused
    elsewhere.

    Stolen from beetsplug.lyrics.
    """
    return re.sub(r'\W+', '-', unidecode(text).lower().strip()).strip('-')


def ensure_directory(path):
    directory = os.path.dirname(path)
    if not os.path.exists(directory):
        os.makedirs(directory)



# -*- coding: utf-8 -*-
import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()

requires = [

    # Foundation
    'six==1.11.0',
    'docopt==0.6.2',
    'tqdm==4.28.1',
    'unidecode==1.0.23',

    # Grafana control and animation
    'where==1.0.2',
    'marionette_driver==2.7.0',
    'python-dateutil==2.7.5',
    #'Pillow==5.2.0',

]

extras = {
    'test': [],
}

setup(name='grafanimate',
      version='0.4.1',
      description='Animate timeseries data with Grafana',
      long_description=README,
      license="AGPL 3, EUPL 1.2",
      classifiers=[
        "Programming Language :: Python",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "License :: OSI Approved :: European Union Public Licence 1.2 (EUPL 1.2)",
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Manufacturing",
        "Intended Audience :: Science/Research",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Telecommunications Industry",
        "Topic :: Communications",
        "Topic :: Database",
        "Topic :: Internet",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Software Development :: Embedded Systems",
        "Topic :: Software Development :: Libraries",
        "Topic :: System :: Archiving",
        "Topic :: System :: Networking :: Monitoring",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS"
        ],
      author='Andreas Motl',
      author_email='andreas@hiveeyes.org',
      url='https://github.com/daq-tools/grafanimate',
      keywords='grafana animate animation automation time gif video',
      packages=find_packages(),
      include_package_data=True,
      package_data={
        'grafanimate': [
          '*.js',
        ],
      },
      zip_safe=False,
      test_suite='grafanimate.test',
      install_requires=requires,
      extras_require=extras,
      tests_require=extras['test'],
      dependency_links=[
      ],
      entry_points={
          'console_scripts': [
              'grafanimate = grafanimate.commands:run',
          ],
      },
)
